// gcc chol.c -o chol.x -Wall -L/Users/pauldj/works/libs/OpenBLAS-develop/ -lopenblas
// gcc chol.c -o chol.x -Wall -L/opt/local/lib/ -lopenblas 

#include <stdlib.h>   // always necessary
#include <stdio.h>    // always necessary
#include <time.h>     // for time()
#include <math.h>     // for sqrt()

#define min(a, b) (((a) < (b)) ? (a) : (b))


int dtrsm_(char *, char *, char *, char *, int *, int *,
	   double *, double *, int *, double *, int *);

int dgemm_(char *, char *, int *, int *, int *, double *,
	   double *, int *, double *, int *, double *, double *, int *);

void chol( double *A, int n, int ldA, int b );
void mat_print( double *A, int n, int m, int ldA, char *text );


int main( int argc, char *argv[] )
{
   int n,i,j;
   double *L, *A;
   double tmp;

   if(argc < 2) 
      { printf(" Arguments needed!\n"); return(-1); }
   else
      n = atoi(argv[1]);
   
   L = (double *) malloc( n*n * sizeof(double) );
   A = (double *) malloc( n*n * sizeof(double) );
   
   srand48( (unsigned)time((time_t *)NULL) ); 
 
   // MATRIX INIT
   for( i=0; i<n; i++ )
      for( j=0; j<=i; j++ )
	 {  tmp = drand48() + drand48(); 
	    A[i+j*n] = tmp;
	    A[j+i*n] = tmp;
	    if(i==j) A[i+i*n] += n; 
	 }

   for( i=0; i<n*n; i++ ) L[i] = A[i];

   chol( L, n, n, 2 );

   return 0;
}



void chol( double *A, int size, int ldA, int b )
{
   int idx=0, blk=0;
   
   mat_print( A, size, size, ldA, "A" );

   if( size == 1 ) { A[0] = sqrt(A[0]); return; }

   while(idx < size)
      {
   	 blk = min( b, size-idx );

   	 if( idx > 0 ){
	    double alpha = 1.0;
	    dtrsm_( "R", "L", "T", "N", &blk, &idx, &alpha, &A[0], &ldA, &A[idx], &ldA );
	    //mat_print( &A[ idx ], blk, idx, ldA, "A10" );
	       
	    alpha = -1;
	    double beta = 1.0;

	    dgemm_( "N", "T", &blk, &blk, &idx, &alpha, &A[idx], &ldA, &A[idx], &ldA, &beta, &A[idx+idx*ldA], &ldA );
	    //mat_print( &A[ idx+idx*ldA ], blk, idx, ldA, "A11" );
   	 }

   	 chol( &A[ idx + idx*ldA ], blk, ldA, 1 );
	 //mat_print( &A[ idx + idx*ldA ], blk, blk, ldA, "A11" );

   	 idx += blk;
      }

   mat_print( A, size, size, ldA, "L" );

   return;
}




void mat_print( double *L, int n, int m, int ldA, char *text )
{
   FILE *output;
   output = stdout;
   int i,j;

   fprintf(output, "%s = [ ...\n", text );
   for( i=0; i<n; i++ ) //row index
      {
	 for( j=0; j<m; j++ )  //column index
	    fprintf(output, "%.6g ", L[i+j*ldA] );
	 fprintf(output, "; ...\n" );
      }
   fprintf(output, "];\n" );

   return;
}
